@@ -1,9 +1,12 @@
 ;;; w3m-hist.el --- a history management system for w3m
-;; Copyright (C) 2001 TSUCHIYA Masatoshi <tsuchiya@pine.kuee.kyoto-u.ac.jp>
+
+;; Copyright (C) 2001 TSUCHIYA Masatoshi <tsuchiya@namazu.org>
 
 ;; Author: Katsumi Yamaoka <yamaoka@jpl.org>
 ;; Keywords: w3m, WWW, hypermedia
 
+;; This file is a part of emacs-w3m.
+
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
@@ -15,15 +18,16 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with this program; if not, you can either send email to this
+;; program's maintainer or write to: The Free Software Foundation,
+;; Inc.; 59 Temple Place, Suite 330; Boston, MA 02111-1307, USA.
 
 ;;; Commentary:
 
-;; w3m keeps history in the buffer-local variables `w3m-history' and
-;; `w3m-history-flat'.  Each variable contains a list of all the links
-;; you have visited.  See the documentations for them for details.
+;; emacs-w3m keeps history in the buffer-local variables `w3m-history'
+;; and `w3m-history-flat'.  Each variable contains a list of all the
+;; links you have visited.  See the documentations for them for
+;; details.
 
 ;;; Code:
 
@@ -231,14 +235,14 @@ will not contain keyword-value pairs whose value is nil."
 	       (while rest
 		 (when (cadr rest)
 		   (setq properties (cons (cadr rest)
-					  (cons (car rest) properties))
-			 rest (cddr rest))))
+					  (cons (car rest) properties))))
+		 (setq rest (cddr rest)))
 	       (nreverse properties))
 	   (while properties
 	     (when (cadr properties)
 	       (setq rest (cons (cadr properties)
-				(cons (car properties) rest))
-		     properties (cddr properties))))
+				(cons (car properties) rest))))
+	     (setq properties (cddr properties)))
 	   (nreverse rest)))))
 
   (defmacro w3m-history-with-element (url set-current
@@ -265,43 +269,33 @@ an error with ERROR-MESSAGE-IF-NO-ELEMENTS."
   (put 'w3m-history-with-element 'lisp-indent-function 'defun))
 
 ;; Functions for internal use.
-(defun w3m-history-copy-1 (tree)
-  "Function used to be called by `w3m-history-copy' and itself
-recursively."
-  (let (element rest branches copy)
-    (while tree
-      (setq element (car tree)
-	    tree (cdr tree)
-	    rest (list (cadr element) (car element))
-	    branches (cddr element))
-      (while branches
-	(setq rest (cons (w3m-history-copy-1 (car branches)) rest)
-	      branches (cdr branches)))
-      (setq copy (cons (nreverse rest) copy)))
-    (nreverse copy)))
-
 (defun w3m-history-seek-properties (url)
   "Seek properties corresponding to URL in all the w3m buffers except for
 the current-buffer."
-  (let ((buffers (delq (current-buffer) (buffer-list)))
-	properties)
+  (let* ((current (current-buffer))
+	 (buffers (delq current (buffer-list)))
+	 properties)
     (while (and (not properties)
 		buffers)
-      (with-current-buffer (pop buffers)
-	(when (eq major-mode 'w3m-mode)
-	  (setq properties (cadr (w3m-history-assoc url))))))
+      (set-buffer (pop buffers))
+      (when (eq major-mode 'w3m-mode)
+	(setq properties (cadr (w3m-history-assoc url)))))
+    (set-buffer current)
     properties))
 
 (defun w3m-history-share-properties (url properties)
   "Function used to keep properties of each history element to be shared
 by all the w3m buffers."
-  (let (flat)
-    (dolist (buffer (buffer-list))
-      (with-current-buffer buffer
-	(when (and (eq major-mode 'w3m-mode)
-		   (setq flat (w3m-history-assoc url)))
-	  (setcar (cdr flat) properties)
-	  (setcar (cdr (w3m-history-current-1 (caddr flat))) properties))))))
+  (let ((buffers (buffer-list))
+	(current (current-buffer))
+	flat)
+    (while buffers
+      (set-buffer (pop buffers))
+      (when (and (eq major-mode 'w3m-mode)
+		 (setq flat (w3m-history-assoc url)))
+	(setcar (cdr flat) properties)
+	(setcar (cdr (w3m-history-current-1 (caddr flat))) properties)))
+    (set-buffer current)))
 
 ;; Generic functions.
 (defun w3m-history-previous-link-available-p ()
@@ -361,35 +355,77 @@ non-nil."
 			 (< (setq count (1+ count)) 0)))))
       (w3m-history-current-1 (cadar w3m-history)))))
 
-(defun w3m-history-flat (&optional history position alist)
+(defun w3m-history-flat ()
   "Set the buffer-local variable `w3m-history-flat' with the value of a
 flattened alist of `w3m-history'.  See the documentation for the
-variable `w3m-history-flat' for details.  Note that the optional
-arguments should only be used to recursive funcall itself internally,
-so don't specify them for the normal use."
-  (if (or history
-	  (setq history (cdr w3m-history)))
-      (progn
-	(unless position
-	  (setq position '(t)))
-	(let ((i 0)
-	      element branches j)
-	  (while (setq element (pop history))
-	    (setcar (nthcdr (1- (length position)) position) i)
-	    (setq i (1+ i))
-	    (push (list (car element) (cadr element) (copy-sequence position))
-		  alist)
-	    (when (setq branches (nthcdr 2 element))
-	      (setq j 0)
-	      (while branches
-		(setq alist (w3m-history-flat (pop branches)
-					      (append position (list j t))
-					      alist)
-		      j (1+ j)))))
-	  (if (cdr position)
-	      alist
-	    (setq w3m-history-flat (nreverse alist)))))
-    (setq w3m-history-flat nil)))
+variable `w3m-history-flat' for details."
+  (setq w3m-history-flat nil)
+  (when w3m-history
+    (let ((history (cdr w3m-history))
+	  (position (list 0))
+	  element branches flag children)
+      (while (setq element (pop history))
+	(if (stringp (car element))
+	    (progn
+	      (push (list (car element) (cadr element) (reverse position))
+		    w3m-history-flat)
+	      (if (setq element (cddr element))
+		  (progn
+		    (setq history (append element history)
+			  position (append (list 0 0) position))
+		    (push (length element) branches))
+		(setcar position (1+ (car position)))
+		(setq flag t)
+		(while (and flag
+			    children
+			    (zerop (setcar children (1- (car children)))))
+		  (setq children (cdr children))
+		  (if (zerop (setcar branches (1- (car branches))))
+		      (progn
+			(setq branches (cdr branches)
+			      position (cddr position))
+			(setcar position (1+ (car position))))
+		    (setcar position 0)
+		    (setcar (cdr position) (1+ (cadr position)))
+		    (setq flag nil)))))
+	  (setq history (append element history))
+	  (push (length element) children))))
+    (setq w3m-history-flat (nreverse w3m-history-flat))))
+
+(defun w3m-history-tree (&optional newpos)
+  "Make a tree-structured history in the variable `w3m-history' from the
+value of `w3m-history-flat'.  The optional NEWPOS should be a list
+of pointers which will be the `car' of the new value of `w3m-history'
+if it is specified.  It defaults to the beginning of a history."
+  (if w3m-history-flat
+      (let ((flat w3m-history-flat)
+	    element positions rest position)
+	(setq w3m-history (list (list nil nil)))
+	(while (setq element (pop flat))
+	  (setq positions (caddr element)
+		rest w3m-history)
+	  (while positions
+	    (setq position (pop positions))
+	    (unless (> (length rest) position)
+	      (setcdr (nthcdr (1- (length rest)) rest)
+		      (make-list (- position (length rest) -1)
+				 (list nil nil))))
+	    (setq rest (nth position rest))
+	    (when positions
+	      (setq position (pop positions))
+	      (unless (> (- (length rest) 2) position)
+		(setcdr (nthcdr (1- (length rest)) rest)
+			(make-list (- position (length rest) -3)
+				   (list (list nil nil)))))
+	      (setq rest (nth (+ position 2) rest))))
+	  (setcar rest (car element))
+	  (setcar (cdr rest) (cadr element)))
+	(if newpos
+	    (setq w3m-history (cons newpos w3m-history))
+	  (setq w3m-history (cons (list nil nil (list 0)) w3m-history))
+	  (w3m-history-forward-1))
+	w3m-history)
+    (setq w3m-history nil)))
 
 (defun w3m-history-assoc (url &optional set-current properties replace-props)
   "Return a history element if URL is `equal' to the `car' of an element
@@ -472,23 +508,23 @@ properties.  See the documentation for the variables `w3m-history' and
     (cond ((null w3m-history)
 	   ;; The dawn of the history.
 	   (let ((shared-plist (w3m-history-seek-properties url)))
+	     (setq properties (w3m-history-modify-properties
+			       shared-plist properties replace-props))
 	     (when shared-plist
-	       (setq properties (w3m-history-modify-properties
-				 shared-plist properties replace-props))
 	       (w3m-history-share-properties url properties)))
 	   (setq element (list url properties)
 		 position '(nil (0) nil)
 		 w3m-history (list position element)
 		 w3m-history-flat (list (append element '((0)))))
 	   position)
-	  ((w3m-history-assoc url t properties)
+	  ((w3m-history-assoc url t properties replace-props)
 	   ;; URL has been registered in the history.
 	   (car w3m-history))
 	  (t
 	   (let ((shared-plist (w3m-history-seek-properties url)))
+	     (setq properties (w3m-history-modify-properties
+			       shared-plist properties replace-props))
 	     (when shared-plist
-	       (setq properties (w3m-history-modify-properties
-				 shared-plist properties replace-props))
 	       (w3m-history-share-properties url properties)))
 	   (let* ((position (copy-sequence (cadar w3m-history)))
 		  (class (1- (length position)))
@@ -521,23 +557,25 @@ properties.  See the documentation for the variables `w3m-history' and
 from BUFFER to the current buffer.  This function keeps properties of
 each history element to be shared between BUFFER and the current
 buffer."
-  (let (flat position tree)
-    (with-current-buffer buffer
-      (setq flat w3m-history-flat
-	    position (car w3m-history)
-	    tree (cons (list (copy-sequence (car position))
-			     (copy-sequence (cadr position))
-			     (copy-sequence (caddr position)))
-		       (w3m-history-copy-1 (cdr w3m-history)))))
-    ;; Remove buffer-local properties.
-    (let (element rest)
-      (while flat
-	(setq element (copy-sequence (car flat))
-	      flat (cdr flat))
-	(setcdr (cddr element) nil)
-	(push element rest))
-      (setq w3m-history-flat (nreverse rest)))
-    (setq w3m-history tree)))
+  (let ((current (current-buffer))
+	position flat)
+    (set-buffer buffer)
+    (when w3m-history
+      (setq position (list (copy-sequence (caar w3m-history))
+			   (copy-sequence (cadar w3m-history))
+			   (copy-sequence (caddar w3m-history)))
+	    flat w3m-history-flat))
+    (set-buffer current)
+    (when position
+      ;; Remove buffer-local properties from the new `w3m-history-flat'.
+      (let (element rest)
+	(while flat
+	  (setq element (copy-sequence (car flat))
+		flat (cdr flat))
+	  (setcdr (cddr element) nil)
+	  (push element rest))
+	(setq w3m-history-flat (nreverse rest)))
+      (w3m-history-tree position))))
 
 (defun w3m-history-plist-get (keyword &optional url set-current local)
   "Extract a value from the properties of a history element.  KEYWORD is
@@ -555,8 +593,7 @@ whether to access the buffer-local properties."
 	(plist-get (if local
 		       (cadddr element)
 		     (cadr element))
-		   keyword)
-      (error "No history element found to be extracted."))))
+		   keyword))))
 
 (defun w3m-history-plist-put (keyword value &optional url set-current local)
   "Change value in the properties of a history element of KEYWORD to
@@ -567,7 +604,7 @@ element corresponding to URL will be set as the current history
 element.  LOCAL specifies whether to access the buffer-local
 properties."
   (w3m-history-with-element url set-current
-    "No history element found to be modified."
+    "No history element found to be modified"
     (if local
 	(let ((properties (w3m-history-modify-properties
 			   nil
@@ -595,7 +632,7 @@ element.  If SET-CURRENT is non-nil, the history element corresponding
 to URL will be set as the current history element.  LOCAL specifies
 whether to access the buffer-local properties."
   (w3m-history-with-element url set-current
-    "No history element found to add properties."
+    "No history element found to add properties"
     (let* ((former (if local
 		       (cadddr element)
 		     (cadr element)))
@@ -639,7 +676,7 @@ element.  If SET-CURRENT is non-nil, the history element corresponding
 to URL will be set as the current history element.  LOCAL specifies
 whether to access the buffer-local properties."
   (w3m-history-with-element url set-current
-    "No history element found to remove properties."
+    "No history element found to remove properties"
     (let* ((former (if local
 		       (cadddr element)
 		     (cadr element)))
@@ -680,70 +717,81 @@ If SET-CURRENT is non-nil, the history element corresponding to URL
 will be set as the current history element.  If THIS-BUFFER is nil,
 renaming will be done for all the w3m buffers."
   (w3m-history-with-element old-url set-current
-    "No history element found to be renamed."
+    "No history element found to be renamed"
     (setcar element new-url)
     (setcar (w3m-history-current-1 (caddr element)) new-url))
   (unless this-buffer
-    (let (element)
-      (dolist (buffer (delq (current-buffer) (buffer-list)))
-	(with-current-buffer buffer
-	  (when (and (eq major-mode 'w3m-mode)
-		     (setq element (w3m-history-assoc old-url)))
-	    (setcar element new-url)
-	    (setcar (w3m-history-current-1 (caddr element)) new-url)))))))
+    (let* ((current (current-buffer))
+	   (buffers (delq current (buffer-list)))
+	   element)
+      (while buffers
+	(set-buffer (pop buffers))
+	(when (and (eq major-mode 'w3m-mode)
+		   (setq element (w3m-history-assoc old-url)))
+	  (setcar element new-url)
+	  (setcar (w3m-history-current-1 (caddr element)) new-url)))
+      (set-buffer current))))
 
-(defun w3m-history-store-position ()
-  "Store the current position point in the history structure."
+(defun w3m-history-store-position (&optional url)
+  "Store the current cursor position in the history structure.  URL
+defaults to the current link."
   (interactive)
   (when (cadar w3m-history)
-    (w3m-history-add-properties (list ':window-start (window-start)
-				      ':position (point))
-				nil nil t)
+    (w3m-history-add-properties (list :window-start (window-start)
+				      :position (point))
+				url nil t)
     (when (interactive-p)
-      (message "The position point has registered"))))
+      (message "The current cursor position has registered"))))
 
-(defun w3m-history-restore-position ()
-  "Restore the last position point for the current url."
+(defun w3m-history-restore-position (&optional url)
+  "Restore the saved cursor position for the page.  URL defaults to the
+current link."
   (interactive)
   (when (cadar w3m-history)
-    (let ((start (w3m-history-plist-get ':window-start nil nil t))
+    (let ((start (w3m-history-plist-get :window-start url nil t))
 	  position)
       (cond ((and start
-		  (setq position (w3m-history-plist-get ':position
-							nil nil t)))
+		  (setq position (w3m-history-plist-get :position
+							url nil t)))
 	     (set-window-start nil start)
 	     (goto-char position))
 	    ((interactive-p)
-	     (message "No last position registered"))))))
+	     (message "No cursor position registered"))))))
 
 (eval-when-compile
   (defvar w3m-arrived-db)
   (autoload 'w3m-goto-url "w3m"))
 
-(defun w3m-history-add-arrived-db ()
+(defun w3m-history-add-arrived-db (&optional clear-history)
   "Add the arrived database to the history structure unreasonably.
-It's only a joke, you should NEVER use it."
-  (interactive)
+If arg CLEAR-HISTORY is given, the current history will be cleared
+in advance.  It's only a joke, you should NEVER use it."
+  (interactive "P")
   (unless (eq 'w3m-mode major-mode)
-    (error "`%s' must be invoked from a w3m buffer." this-command))
+    (error "`%s' must be invoked from a w3m buffer" this-command))
   (when (and w3m-arrived-db
 	     (prog1
 		 (yes-or-no-p
 		  "Are you sure you really want to destroy the history? ")
 	       (message "")))
-    (when w3m-history
-      (w3m-history-push (car (w3m-history-current-1 '(0)))))
-    (let (url title)
-      (mapatoms
-       (function
-	(lambda (symbol)
-	  (when symbol
-	    (setq url (symbol-name symbol)
-		  title (get symbol 'title))
-	    (w3m-history-push url (when title
-				    (list ':title title)))
-	    (w3m-history-backward (random 2) t))))
-       w3m-arrived-db))
+    (when clear-history
+      (setq w3m-history nil
+	    w3m-history-flat nil))
+    (let (url-title title)
+      (mapatoms (function
+		 (lambda (symbol)
+		   (when symbol
+		     (if (setq title (get symbol 'title))
+			 (push (list (symbol-name symbol)
+				     (list :title title))
+			       url-title)
+		       (push (list (symbol-name symbol)) url-title)))))
+		w3m-arrived-db)
+      (apply 'w3m-history-push (nth (random (length url-title)) url-title))
+      (while url-title
+	(w3m-history-push (car (nth (random (length w3m-history-flat))
+				    w3m-history-flat)))
+	(apply 'w3m-history-push (pop url-title))))
     (w3m-goto-url "about://history/")))
 
 (provide 'w3m-hist)
